/********************************************************************
 Software License Agreement:

 The software supplied herewith by Microchip Technology Incorporated
 (the "Company") for its PIC(R) Microcontroller is intended and
 supplied to you, the Company's customer, for use solely and
 exclusively on Microchip PIC Microcontroller products. The
 software is owned by the Company and/or its supplier, and is
 protected under applicable copyright laws. All rights are reserved.
 Any use in violation of the foregoing restrictions may subject the
 user to criminal sanctions under applicable laws, as well as to
 civil liability for the breach of the terms and conditions of this
 license.

 THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES,
 WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *******************************************************************/

/** INCLUDES *******************************************************/
#include "system.h"
#include "system_config.h"

#include "usb.h"
#include "usb_device.h"
#include "usb_device_cdc.h"

#define LED_ACTIVITY_MS  25

/** VARIABLES SERIAL 1 *********************************************/

static uint8_t USB_Out_Buffer0[CDC_DATA_OUT_EP_SIZE];
static uint8_t RS232_Out_Data0[CDC_DATA_IN_EP_SIZE];

static unsigned char  NextUSBOut0;
static unsigned char  LastRS232Out0;  // Number of characters in the buffer
static unsigned char  RS232cp0;       // current position within the buffer
static unsigned char  RS232_Out_Data_Rdy0;

static uint8_t USART_In_Buffer0[CDC_DATA_OUT_EP_SIZE];
static unsigned char  In_Buffer_Head0;
static unsigned char  In_Buffer_Tail0;

static uint16_t ledCount0;

/** VARIABLES SERIAL 2 *********************************************/

static uint8_t USB_Out_Buffer1[CDC_DATA_OUT_EP_SIZE];
static uint8_t RS232_Out_Data1[CDC_DATA_IN_EP_SIZE];

static unsigned char  NextUSBOut1;
static unsigned char  LastRS232Out1;  // Number of characters in the buffer
static unsigned char  RS232cp1;       // current position within the buffer
static unsigned char  RS232_Out_Data_Rdy1;

static uint8_t USART_In_Buffer1[CDC_DATA_OUT_EP_SIZE];
static unsigned char  In_Buffer_Head1;
static unsigned char  In_Buffer_Tail1;

static uint16_t ledCount1;

/********************************************************************
 * Function:        void main(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Main program entry point.
 *
 * Note:            None
 *******************************************************************/
MAIN_RETURN main(void)
{
    // Map peripherals
    EECON2 = 0x55; EECON2 = 0xAA; PPSCONbits.IOLOCK = 0;
    RPINR16 = 13; // Map RP13/RC2 to USART2 RX
    RPOR12 = 5;   // Map RP12/RC1 to USART2 TX
    EECON2 = 0x55; EECON2 = 0xAA; PPSCONbits.IOLOCK = 1;

    // Configure all pins as digital ports
    ANCON0 = 0x1F;
    ANCON1 = 0x1F;
    ODCON1 = ODCON2 = ODCON3 = 0x00;
    LATA = LATB = LATC = 0;

    //In this devices family of USB microcontrollers, the PLL will not power up and be enabled
    //by default, even if a PLL enabled oscillator configuration is selected (such as HS+PLL).
    //This allows the device to power up at a lower initial operating frequency, which can be
    //advantageous when powered from a source which is not gauranteed to be adequate for 48MHz
    //operation.  On these devices, user firmware needs to manually set the OSCTUNE<PLLEN> bit to
    //power up the PLL.
    {
        unsigned int pll_startup_counter = 600;
        OSCTUNEbits.PLLEN = 1;  //Enable the PLL and wait 2+ms until the PLL locks before enabling USB module
        while(pll_startup_counter--);
    }
    //Device switches over automatically to PLL output after PLL is locked and ready.

    LED_Enable(LED_USB_DEVICE_STATE);
    LED_Enable(LED_USB_SERIAL1_ACTIVITY);
    LED_Enable(LED_USB_SERIAL2_ACTIVITY);

	NextUSBOut0 = NextUSBOut1 = 0;
	LastRS232Out0 = LastRS232Out1 = 0;
	RS232_Out_Data_Rdy0 = RS232_Out_Data_Rdy1 = 0;
    In_Buffer_Head0 = In_Buffer_Tail0 = 0;
    In_Buffer_Head1 = In_Buffer_Tail1 = 0;
    ledCount0 = ledCount1 = 0;

    USART_Initialize();
    USBDeviceInit();
    USBDeviceAttach();
    
    while(1)
    {
        /* If the USB device isn't configured yet, we can't really do anything
         * else since we don't have a host to talk to.  So jump back to the
         * top of the while loop. */
        if( USBGetDeviceState() < CONFIGURED_STATE )
        {
            /* Jump back to the top of the while loop. */
            continue;
        }

        /* If we are currently suspended, then we need to see if we need to
         * issue a remote wakeup.  In either case, we shouldn't process any
         * keyboard commands since we aren't currently communicating to the host
         * thus just continue back to the start of the while loop. */
        if( USBIsDeviceSuspended() == true )
        {
            /* Jump back to the top of the while loop. */
            continue;
        }

        /* Serial 1 ****************************************************/

        if (RS232_Out_Data_Rdy0 == 0)  // only check for new USB buffer if the old RS232 buffer is
        {						      // empty.  This will cause additional USB packets to be NAK'd
            LastRS232Out0 = getsUSBUSART(0,RS232_Out_Data0,CDC_DATA_IN_EP_SIZE); //until the buffer is free.
            if(LastRS232Out0 > 0)
            {
                RS232_Out_Data_Rdy0 = 1;  // signal buffer full
                RS232cp0 = 0;  // Reset the current position
                ledCount0 = LED_ACTIVITY_MS;
                LED_On(LED_USB_SERIAL1_ACTIVITY);
            }
        }

        // Check if one or more bytes are waiting in the physical UART transmit
        // queue.  If so, send it out the UART TX pin.
        if (RS232_Out_Data_Rdy0 && UART1_TxRdy())
        {
            USART_putcUSART1(RS232_Out_Data0[RS232cp0]);
            ++RS232cp0;
            if (RS232cp0 == LastRS232Out0)
                RS232_Out_Data_Rdy0 = 0;
        }

        // Check if any bytes are waiting in the queue to send to the USB host.
        // If any bytes are waiting, and the endpoint is available, prepare to
        // send the USB packet to the host.
        if (USBUSARTIsTxTrfReady(0) && In_Buffer_Head0 != In_Buffer_Tail0)
        {
            NextUSBOut0 = 0;
            do {
                USB_Out_Buffer0[NextUSBOut0] = USART_In_Buffer0[In_Buffer_Tail0];
                ++NextUSBOut0;
                ++In_Buffer_Tail0;
                if (In_Buffer_Tail0 >= sizeof(USB_Out_Buffer0))
                    In_Buffer_Tail0 = 0;
            } while (In_Buffer_Head0 != In_Buffer_Tail0 && NextUSBOut0 < CDC_DATA_OUT_EP_SIZE);
            putUSBUSART(0, &USB_Out_Buffer0[0], NextUSBOut0);
            ledCount0 = LED_ACTIVITY_MS;
            LED_On(LED_USB_SERIAL1_ACTIVITY);
        }

        /* Serial 2 ****************************************************/

        if (RS232_Out_Data_Rdy1 == 0)  // only check for new USB buffer if the old RS232 buffer is
        {						      // empty.  This will cause additional USB packets to be NAK'd
            LastRS232Out1 = getsUSBUSART(1,RS232_Out_Data1,CDC_DATA_IN_EP_SIZE); //until the buffer is free.
            if(LastRS232Out1 > 0)
            {
                RS232_Out_Data_Rdy1 = 1;  // signal buffer full
                RS232cp1 = 0;  // Reset the current position
                ledCount1 = LED_ACTIVITY_MS;
                LED_On(LED_USB_SERIAL2_ACTIVITY);
            }
        }

        // Check if one or more bytes are waiting in the physical UART transmit
        // queue.  If so, send it out the UART TX pin.
        if (RS232_Out_Data_Rdy1 && UART2_TxRdy())
        {
            USART_putcUSART2(RS232_Out_Data1[RS232cp1]);
            ++RS232cp1;
            if (RS232cp1 == LastRS232Out1)
                RS232_Out_Data_Rdy1 = 0;
        }

        // Check if any bytes are waiting in the queue to send to the USB host.
        // If any bytes are waiting, and the endpoint is available, prepare to
        // send the USB packet to the host.
        if (USBUSARTIsTxTrfReady(1) && In_Buffer_Head1 != In_Buffer_Tail1)
        {
            NextUSBOut1 = 0;
            do {
                USB_Out_Buffer1[NextUSBOut1] = USART_In_Buffer1[In_Buffer_Tail1];
                ++NextUSBOut1;
                ++In_Buffer_Tail1;
                if (In_Buffer_Tail1 >= sizeof(USB_Out_Buffer1))
                    In_Buffer_Tail1 = 0;
            } while (In_Buffer_Head1 != In_Buffer_Tail1 && NextUSBOut1 < CDC_DATA_OUT_EP_SIZE);
            putUSBUSART(1, &USB_Out_Buffer1[0], NextUSBOut1);
            ledCount1 = LED_ACTIVITY_MS;
            LED_On(LED_USB_SERIAL2_ACTIVITY);
        }

        USBMaskInterrupts();

        CDCTxService(0);
        CDCTxService(1);

        USBUnmaskInterrupts();

    }//end while
}//end main

bool USER_USB_CALLBACK_EVENT_HANDLER(USB_EVENT event, void *pdata, uint16_t size)
{
    switch( (int) event )
    {
        case EVENT_TRANSFER:
            break;

        case EVENT_SOF:
            /* We are using the SOF as a timer to time the LED indicator.  Call
             * the LED update function here. */
            if (ledCount0 > 0) {
                ledCount0--;
                if (ledCount0 == 0)
                    LED_Off(LED_USB_SERIAL1_ACTIVITY);
            }
            if (ledCount1 > 0) {
                ledCount1--;
                if (ledCount1 == 0)
                    LED_Off(LED_USB_SERIAL2_ACTIVITY);
            }
            break;

        case EVENT_SUSPEND:
            /* Update the LED status for the suspend event. */
            LED_Off(LED_USB_DEVICE_STATE);
            LED_Off(LED_USB_SERIAL1_ACTIVITY);
            LED_Off(LED_USB_SERIAL2_ACTIVITY);
            break;

        case EVENT_RESUME:
            OSCCON = 0x70;		//Primary clock source selected.
            
            //Adding a software start up delay will ensure
            //that the primary oscillator and PLL are running before executing any other
            //code.  If the PLL isn't being used, (ex: primary osc = 48MHz externally applied EC)
            //then this code adds a small unnecessary delay, but it is harmless to execute anyway.
            {
                unsigned int pll_startup_counter = 800;	//Long delay at 31kHz, but ~0.8ms at 48MHz
                while(pll_startup_counter--);			//Clock will switch over while executing this delay loop
            }
            /* Update the LED status for the resume event. */
            LED_On(LED_USB_DEVICE_STATE);
            break;

        case EVENT_CONFIGURED:
            CDCInitEP();
            LED_On(LED_USB_DEVICE_STATE);
            break;

        case EVENT_SET_DESCRIPTOR:
            break;

        case EVENT_EP0_REQUEST:
            /* We have received a non-standard USB request.  The HID driver
             * needs to check to see if the request was for it. */
            USBCheckCDCRequest();
            break;

        case EVENT_BUS_ERROR:
            break;

        case EVENT_TRANSFER_TERMINATED:
            break;

        default:
            break;
    }
    return true;
}

#if defined(__XC8)
void interrupt SYS_InterruptHigh(void)
{
    // Check if we received a character over the physical UARTs, and we need
    // to buffer it up for eventual transmission to the USB host.

    if (UART1_DataRdy())
    {
        USART_In_Buffer0[In_Buffer_Head0] = USART_getcUSART1();
        ++In_Buffer_Head0;
        if (In_Buffer_Head0 >= sizeof(USART_In_Buffer0))
            In_Buffer_Head0 = 0;
    }

    if (UART2_DataRdy())
    {
        USART_In_Buffer1[In_Buffer_Head1] = USART_getcUSART2();
        ++In_Buffer_Head1;
        if (In_Buffer_Head1 >= sizeof(USART_In_Buffer1))
            In_Buffer_Head1 = 0;
    }

    #if defined(USB_INTERRUPT)
        USBDeviceTasks();
    #endif
}
#else
    void YourHighPriorityISRCode();
    void YourLowPriorityISRCode();

    //On PIC18 devices, addresses 0x00, 0x08, and 0x18 are used for
    //the reset, high priority interrupt, and low priority interrupt
    //vectors.  However, the current Microchip USB bootloader
    //examples are intended to occupy addresses 0x00-0x7FF or
    //0x00-0xFFF depending on which bootloader is used.  Therefore,
    //the bootloader code remaps these vectors to new locations
    //as indicated below.  This remapping is only necessary if you
    //wish to program the hex file generated from this project with
    //the USB bootloader.  If no bootloader is used, edit the
    //usb_config.h file and comment out the following defines:
    //#define PROGRAMMABLE_WITH_USB_HID_BOOTLOADER
    //#define PROGRAMMABLE_WITH_USB_LEGACY_CUSTOM_CLASS_BOOTLOADER

	#if defined(PROGRAMMABLE_WITH_USB_HID_BOOTLOADER)
		#define REMAPPED_RESET_VECTOR_ADDRESS			0x1000
		#define REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS	0x1008
		#define REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS	0x1018
	#elif defined(PROGRAMMABLE_WITH_USB_MCHPUSB_BOOTLOADER)
		#define REMAPPED_RESET_VECTOR_ADDRESS			0x800
		#define REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS	0x808
		#define REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS	0x818
	#else
		#define REMAPPED_RESET_VECTOR_ADDRESS			0x00
		#define REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS	0x08
		#define REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS	0x18
	#endif

	#if defined(PROGRAMMABLE_WITH_USB_HID_BOOTLOADER)||defined(PROGRAMMABLE_WITH_USB_MCHPUSB_BOOTLOADER)
	extern void _startup (void);        // See c018i.c in your C18 compiler dir
	#pragma code REMAPPED_RESET_VECTOR = REMAPPED_RESET_VECTOR_ADDRESS
	void _reset (void)
	{
	    _asm goto _startup _endasm
	}
	#endif

        #pragma code REMAPPED_HIGH_INTERRUPT_VECTOR = REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS
	void Remapped_High_ISR (void)
	{
	     _asm goto YourHighPriorityISRCode _endasm
	}
	#pragma code REMAPPED_LOW_INTERRUPT_VECTOR = REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS
	void Remapped_Low_ISR (void)
	{
	     _asm goto YourLowPriorityISRCode _endasm
	}

	#if defined(PROGRAMMABLE_WITH_USB_HID_BOOTLOADER)||defined(PROGRAMMABLE_WITH_USB_MCHPUSB_BOOTLOADER)
	//Note: If this project is built while one of the bootloaders has
	//been defined, but then the output hex file is not programmed with
	//the bootloader, addresses 0x08 and 0x18 would end up programmed with 0xFFFF.
	//As a result, if an actual interrupt was enabled and occured, the PC would jump
	//to 0x08 (or 0x18) and would begin executing "0xFFFF" (unprogrammed space).  This
	//executes as nop instructions, but the PC would eventually reach the REMAPPED_RESET_VECTOR_ADDRESS
	//(0x1000 or 0x800, depending upon bootloader), and would execute the "goto _startup".  This
	//would effective reset the application.

	//To fix this situation, we should always deliberately place a
	//"goto REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS" at address 0x08, and a
	//"goto REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS" at address 0x18.  When the output
	//hex file of this project is programmed with the bootloader, these sections do not
	//get bootloaded (as they overlap the bootloader space).  If the output hex file is not
	//programmed using the bootloader, then the below goto instructions do get programmed,
	//and the hex file still works like normal.  The below section is only required to fix this
	//scenario.
	#pragma code HIGH_INTERRUPT_VECTOR = 0x08
	void High_ISR (void)
	{
	     _asm goto REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS _endasm
	}
	#pragma code LOW_INTERRUPT_VECTOR = 0x18
	void Low_ISR (void)
	{
	     _asm goto REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS _endasm
	}
	#endif	//end of "#if defined(PROGRAMMABLE_WITH_USB_HID_BOOTLOADER)||defined(PROGRAMMABLE_WITH_USB_LEGACY_CUSTOM_CLASS_BOOTLOADER)"

	#pragma code


	//These are your actual interrupt handling routines.
	#pragma interrupt YourHighPriorityISRCode
	void YourHighPriorityISRCode()
	{
		//Check which interrupt flag caused the interrupt.
		//Service the interrupt
		//Clear the interrupt flag
		//Etc.
        #if defined(USB_INTERRUPT)
	        USBDeviceTasks();
        #endif

	}	//This return will be a "retfie fast", since this is in a #pragma interrupt section
	#pragma interruptlow YourLowPriorityISRCode
	void YourLowPriorityISRCode()
	{
		//Check which interrupt flag caused the interrupt.
		//Service the interrupt
		//Clear the interrupt flag
		//Etc.

	}	//This return will be a "retfie", since this is in a #pragma interruptlow section
#endif     
